# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains DataSetup class for setting-up, initialization,
configuration, and tearing-down of data Environment
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.2.0"


from typing import Optional, Dict
from pandas import DataFrame as DF

from snowflake_ai.common import DataConnect
from snowflake_ai.connect import FileConnect, \
    SnowConnect, DataFrameFactory



class DataSetup:
    """
    This class sets up, initializes, configures or tears down
    the data Environment

    To use this class, instantiate DataSetup with appropriate DataConnect
    as follows:

        >>> from snowflake_ai.common import FileConnect
        >>> from snowflake_ai.snowpandas import DataSetup
        ... 
        >>> conn: FileConnect = FileConnect()
        >>> setup: DataSetup = DataSetup(conn)
        >>> df = setup.load_data()
    """
    def __init__(
        self, 
        datasetup_key: str,
        connect: DataConnect, 
        data: Optional[Dict[str, DF]] = {}
    ) -> None:
        self.datasetup_key = datasetup_key
        self._connect = connect
        self._data = data
        if data is not None:
            self._data = data


    def get_connect(self) -> DataConnect:
        """
        Get the data connection for the setup.

        Returns:
            DataConnect: data connection object for the setup
        """        
        return self._connect
    

    def load_data(self, file_name: Optional[str] = None) -> DF:
        """
        Load pandas dataframe from local-csv file connection.

        Returns:
            DataFrame: Pandas dataframe
        """
        if isinstance(self._connect, FileConnect):
            if (file_name is None) or (file_name.strip() == ""):
                return DataFrameFactory.create_pdf(
                    "", self._connect
                )
            else:
                df: DF = DataFrameFactory.create_pdf(
                    file_name, self._connect
                )
                
                df.columns = df.columns.str.strip().str.lower()\
                    .str.replace(' ', '_')
                return df
        elif isinstance(self._connect, SnowConnect):
            pass
        else:
            raise ValueError(
                "DataSetup Error. Loading File requires FileConnect"
            )


    def _load_file_data(self):
        pass


    def _load_snowflake_data(self):
        pass


    def get_data(self) -> Dict[str, DF]:
        """
        Get all dataframes for exploration and preparation.

        Returns:
            Dict: dictionary of dataframes
        """
        return self._data


    def clean_data(self, file_name: Optional[str] = None):
        self._data = {}
        self._connect.close_connection()


# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains StreamlitApp class representing an application
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.2.0"


import os
from os.path import exists
import sys
from typing import Optional, Union, Dict, Callable
import logging
import toml
from functools import wraps
import streamlit as st

from snowflake_ai.common import AppConfig
from snowflake_ai.apps import AppPage
from snowflake_ai.common import OAuthConnect
from snowflake_ai.connect import AuthCodeConnect, ConnectManager



class StreamlitApp(AppConfig):
    """
    This class represents a Steamlit application's configurations and 
    its corresponding application. Since it is subclass of AppConfig,
    it has the following config directory bootstrapping precedence:

    1) input custom directory 
    2) snowflake_ai/conf subdir under current directory 
    3) snowflake_ai/conf subdir under user_home directory
    4) current directory
    5) home directory
    6) conf subdir under snowflake_ai library installation root dir

    Assuming a streamlit application app_1 under a default group named as
        streamlit_default is configured, this app can be created as:

        >>> from snowflake_ai.apps import StreamlitApp
        ... 
        ... # initialize application config for app_1
        >>> app = StreamlitApp("streamlit_default.app_1")
        ...
        ... # optionally the configs may be loaded from your custom dir
        ... # app = AppConfig("group_0.app_1", "custom dir", "config.toml")
    """

    _logger = logging.getLogger(__name__)
    _logger.addHandler(logging.StreamHandler(sys.stdout))

    K_PAGE = "page"
    K_PRE_PAGE = "previous_page"
    T_ST_APP = "streamlit"


    def __init__(
        self,
        app_key : str,
        config_dir : Optional[Union[str, None]] = None, 
        config_file : Optional[Union[str, None]] = None
    ):
        """
        Creat app specific AppConfig object.

        Args:
            app_key: A string representing the application; it can have 
                the format of <app_group>.<application> ('<', '>' not 
                included)
            config_dir: directory path string for custom config load
            config_file: file name string for custom config load
        """
        super().__init__(app_key, config_dir, config_file)
        self.logger = AppConfig._logger
        if StreamlitApp.T_ST_APP not in self.type:
            raise TypeError(
                "StreamlitApp.init(): Application type configuration Error."
            )
        self._setup_streamlit_config()
        self.pages : Dict[str, AppPage] = {}
        self.is_logged_in = False


    def _setup_streamlit_config(
        self, 
        conf_file: Optional[str] = None,
        file_content: Optional[str] = None
    ) -> str:
        """
        Setup streamlit app config file.

        Args:
            conf_file (str): config file name
            file_content (str): custom config file content
        
        Retrun:
            str: streamlit configuration file path
        """
        app_dir = os.path.join(self.root_path, self.script_home_dir)
        p = os.path.join(app_dir, ".streamlit/")
        if not exists(p):
            try:
                os.makedirs(p, exist_ok=True)
            except Exception as e:
                raise ValueError(
                    f"Streamlit._setup_streamlit_dir(): Error - {e}. "\
                    "Cannot create streamlit configuration directory"
                )

        if conf_file is None:
            conf_file = "config.toml"
        
        conf_path = os.path.join(p, conf_file)

        if not exists(conf_path):
            if file_content is None:
                file_content = self.get_streamlit_config()
            
            if file_content:
                with open(conf_path, "w") as f:
                    f.write(file_content)
        else:
            with open(conf_path, "r") as f:
                current_content = f.read()
            
            if file_content is None:
                file_content = self.get_streamlit_config()
            
            if current_content != file_content:
                if file_content:
                    with open(conf_path, "w") as f:
                        f.write(file_content)

        return conf_path


    def get_streamlit_config(self) :
        """
        Get streamlit app configuration from config file
        """
        # split app_key in form of group_key.app_key into two variables
        gk, k = AppConfig.split_group_key(self.app_key)
        config: Dict[str, Dict] = self.get_all_configs().get(
            AppConfig.K_ST_APPS, {}
        )
        self.logger.info(
            f"StreamlitApp.get_streamlit_config(): group_key =>{gk}; app_"\
            f"key => {k}; config => {config}"
        )
        if config:
            config = config.get(gk, {})
            if config:
                config = config.get(k, {})
        
        return toml.dumps(config)
    

    def add_page(self, page:AppPage):
        if self.pages:
            self.pages[page.page_id] = page
        

    def _add_page(
        self,
        page_id: str,
        title: Optional[str] = None,
        icon: Optional[str] = None,
        layout: Optional[str] = "wide",
        show_sidebar: Optional[bool] = True,
        func: Optional[Callable] = None,
    ):
        page = AppPage(
            page_id=page_id,
            title=title,
            icon=icon,
            layout=layout,
            show_sidebar=show_sidebar,
            func=func,
        )
        self.pages[page_id] = page


    def register_page(
        self,
        page_id: str,
        title: Optional[str] = None,
        icon: Optional[str] = None,
        layout: Optional[str] = "wide",
        show_sidebar: Optional[bool] = True,
    ) -> Callable:
        def decorator(func: Callable) -> Callable:
            @wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            self._add_page(
                page_id=page_id,
                title=title,
                icon=icon,
                layout=layout,
                show_sidebar=show_sidebar,
                func=wrapper,
            )
            return wrapper

        return decorator
    

    def set_pages_config(self):
        for _, pg in self.pages.items():
            st.set_page_config(
                page_title = pg.title,
                page_icon = pg.icon ,
                layout = pg.layout,
            )


    @staticmethod
    def get_current_page():
        query_params = st.experimental_get_query_params()
        return query_params.get(StreamlitApp.K_PAGE, ["/"])[0]
    

    def navigate_to(self, page_id: str):
        previous_page_id = self.get_current_page()
        st.experimental_set_query_params(
            previous_page = previous_page_id
        )
        st.experimental_set_query_params(page=page_id)
        self.run()


    def run(self):
        current_page_id = self.get_current_page()
        page: AppPage = self.pages.get(current_page_id)
        if page:
            page.func()
        else:
            st.error("Page not found!")


    def request_auth_code(
        self,
        oc:OAuthConnect
    ) -> str:
        oc = ConnectManager.create_default_oauth_connect(self)
        _, cc = AuthCodeConnect.generate_pkce_pair()
        params = {
            AuthCodeConnect.K_CODE_CHALLENGE: cc
        }
        return oc.authorize_request(params)


    def request_access_token(
        self,
        oc: OAuthConnect,
        ap: AppPage = None
    ) -> Dict :
        ok: bool = False
        params = oc.prepare_grant_request()
        auth_cd = params.get("auth_code")
        ctx = {}
        if auth_cd is not None and auth_cd:
            tok_res = oc.grant_request(params)
            print(f"DEBUG S0 ========> token result: {tok_res}")
            if tok_res:
                dc = ConnectManager.create_default_snow_connect(self)
                ctx = oc.decode_token(
                    tok_res, ["access_token", "refresh_token"]
                )
                print(f"DEBUG S1 => context => {ctx.items}")
                session = dc.create_session(ctx)
                if ap is not None:
                    ap.session = session
                    print(f"DEBUG S2 => session: true")

                tok: Dict = ctx.get("decoded_access_token")
                if tok:
                    fnm = tok.get("given_name", "")
                    lnm = tok.get("family_name", "")
                    st.sidebar.write(
                        f"Welcome, {fnm} {lnm}!"
                    )
                ok = True
        if not ok:
            st.error("Authorization failed.")

        return ctx


    def request_refresh_token(
        self,
        ctx: Dict,
        oc: OAuthConnect,
        ap: AppPage = None
    ) -> Dict :
        ok: bool = False
        params: Dict = oc.prepare_token_refresh(ctx) 
        refresh_tok = params.get("refresh_token")
        ctx = {}
        if  refresh_tok is not None and refresh_tok:
            tok_res = oc.refresh_token_request(params)
            print(f"DEBUG S1 ===========> refersh result: {tok_res}")
            if tok_res:
                dc = ConnectManager.create_default_snow_connect(self)
                ctx = oc.decode_token(
                    tok_res, ["access_token", "refresh_token"]
                )
                if ctx.get("access_token"):
                    print(f"DEBUG S2 ===========> context: true")
                session = dc.create_session(ctx)
                if ap is not None:
                    ap.session = session

                tok: Dict = ctx.get("decoded_access_token")
                if tok:
                    fnm = tok.get("given_name", "")
                    lnm = tok.get("family_name", "")
                    st.sidebar.write(
                        f"Welcome, {fnm} {lnm}!"
                    )
                ok = True
        if not ok:
            st.error("Authorization failed.")

        return ctx
    


# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains OAuthConnect class representing a generic OAuth
connection configuration.
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.2.0"


import sys
from typing import Optional, Dict, List
import logging

from snowflake_ai.common import AppConfig
from snowflake_ai.snowpandas import DataSetup



class SetupManager:
    """
    This class represents a generic application connect for 
    OAuth grant flow.

    To use this class, create instance from AppConfig or extend
    it in a child class for creation of a specific OAuth connect
    instance, e.g., AuthCodeConnect as following :

        >>> from snowflake_ai.connect import AuthCodeConnect
        ...
        >>> ac = AppConfig("group_0.app_1")
        >>> conn = SecurityManager(ac).create_default_oauth_connect()
        >>> conn.grant_request()
    """

    _logger = logging.getLogger(__name__)
    _logger.addHandler(logging.StreamHandler(sys.stdout))

    _initialized = False
    _init_connect_params = {}


    def __init__(self, app_conf: AppConfig):
        pass
    

    def create_data_setups(self) -> List[DataSetup]:
        """
        Create list of data setup instances

        Returns:
            List: data setups instances
        """   
        return []




# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains FlowContext class representing the context
of MLOps flow
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.2.0"


import logging



class FlowContext:
    """
    This class provides a context for the MLOps flow consisting
    of direct input (prior processing/task), context input (full
    historical context sequence), current output, and related
    metadata.

    Example:

    To use this class, instantiate the initial context:

        >>> from snowflake_ai.mlops import FlowContext
        ... 
        >>> ctx: FlowContext = FlowContext()
    """

    def __init__(self) -> None:
        self.logger = logging.getLogger(__name__)
        self.direct_input = {}
        self.context_input = []
        self.output = {}
        self.metadata = {}




# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains FlowOps class for ML flow operationization
targeting production domain/environment
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.2.0"


from snowflake_ai.mlops import Pipeline


class FlowOps(Pipeline):
    """
    This class provides general setup for MLops flow in
    operationalizing the model pipeline built, including deloyment
    serving the model, monitor and re-train the model 
    (To-DO).
    """

    def __init__(self) -> None:
        super().__init__()



# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains SnowConnect class representing a specific Snowflake
connection.
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.2.0"


import os
import sys
import logging
import types
from typing import List, Dict, Iterator, Optional

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

from pandas import DataFrame as DF

from snowflake.snowpark import Session
from snowflake.snowpark import Row
from snowflake.snowpark import DataFrame as SDF

from snowflake_ai.common import AppConfig, AppConnect, DataConnect



class SnowConnect(DataConnect):
    """
    This class provides specific Snowflake connection.

    Example:

    To use this class, just instantiate SnowConnect as follows:

        >>> from snowflake_ai.common import SnowConnect
        ... 
        >>> connect: SnowConnect = SnowConnect()
        >>> session = connect.get_session()
    """

    T_SNOWFLAKE_CONN = AppConnect.T_SNOWFLAKE_CONN
    T_AUTH_SNOWFLAKE = AppConnect.T_AUTH_SNOWFLAKE
    T_AUTH_KEYPAIR = AppConnect.T_AUTH_KEYPAIR
    T_AUTH_EXT_BROWSER = AppConnect.T_AUTH_EXT_BROWSER
    T_AUTH_OAUTH = AppConnect.T_AUTH_OAUTH

    _logger = logging.getLogger(__name__)
    _logger.addHandler(logging.StreamHandler(sys.stdout))


    def __init__(self, connect_key : Optional[str] = None):
        super().__init__(connect_key)
        self.logger = SnowConnect._logger
        
        if not (self.connect_group == DataConnect.K_DATA_CONN and \
                self.connect_params):
            self.logger.error(
                f"SnowConnect.init(): ConnectType => {self.connect_type};"\
                f"ConnectGroup => {self.connect_group}; ConnectName => "\
                f"{self.connect_name}"
            )
            raise ValueError(
                "SnowConnect.init(): SnowConnect configuration Error!"
            )
        if (not self.connect_type) or (
            self.connect_type != AppConfig.T_SNOWFLAKE_CONN
        ):
            raise ValueError(
                f"SnowConnect.init(): Type [{self.connect_type}] Error"
            )


    def create_connection(self, params: Dict):
        """
        Create a snowflake connection based on the specific configuration.
        A set of authentication types are supported:
            snowflake, keypair, oauth and externalbrowser
        Note, user specific connection would be session.

        Args:
            params (Dict): Configuration dictionary as input parameters.

        Returns:
            object: snowflake connection session
        """
        if params is None:
            raise ValueError(
                f"SnowConnect.create_connection params cannot be None"
            )
        session = None
        try:
            auth = params[AppConfig.K_AUTH_TYPE]
            if auth == AppConnect.T_AUTH_SNOWFLAKE:
                session = self._do_snowflake_auth(params)
            elif auth == AppConnect.T_AUTH_KEYPAIR:
                session = self._do_keypair_auth(params)
            elif auth == AppConnect.T_AUTH_EXT_BROWSER: 
                # cannot create shared connection
                self.logger.warning(
                    "SnowConnect.create_connection(): use create_session "\
                    "to create externalbrowser enabled connection."
                )
            elif auth == AppConnect.T_AUTH_OAUTH:
                # cannot create shared connection
                self.logger.warning(
                    "SnowConnect.create_connection(): use create_session "\
                    "to create oauth enabled connection."
                )
        except Exception as e:
            self.logger.exception(
                f"SnowConnect.create_connection cannot use "\
                f"auth_type={auth} to creaet snowflake session: {e}"
            )
        return session


    def get_connection(self, connect_key: Optional[str] = None):
        """
        Get lazy created snowflake shared connection.

        Args:
            connect_key (str): data connect key.

        Returns:
            object: snowflake connection object, i.e., snowflake session
        """
        conn = super().get_connection(connect_key)
        if conn is not None and (
            (self.connect_params["type"] != SnowConnect.T_SNOWFLAKE_CONN) \
            or (not isinstance(conn, Session))
        ):
            raise TypeError(
                f"SnowConnect.get_connection(): Type Error {connect_key}"
            )
        return conn


    def is_current_active(self) -> bool:
        """
        Check whether current Snowflake connection (shared session) is
        still active or not. 
        
        Returns:
            bool: True if the connection is active, otherwise False
        """
        rb = super().is_current_active()
        try:
            r = self.dql("select current_role()")           
            try:
                pd = next(r)
                self.logger.info(
                    "SnowConnect.is_current_active(): Active "
                    f"connection using role [{pd}]"
                )
                if DF(pd).shape[0] > 0:
                    rb = True
            except Exception as ee:
                self.logger.warning(
                    "SnowConnect.is_current_active(): No Result "\
                    f"returned [{ee}]"
                )            
        except Exception as e:
            self.logger.warning(
                f"SnowConnect.is_current_active(): Warning {e}"
            )
        return rb
    

    def create_session(self, ctx: Optional[Dict] = {}) -> Session:
        """
        Create snowflake user connection session, or get the existing
        shared data connection.

        Returns:
            Session: Snowflake connection's session
        """
        conn = self.get_connection()
        if (conn is not None) and isinstance(conn, Session):
            return conn
        else:
            conn = self.create_connection(self.connect_params)
        session = conn
        if session is None:
            if self.auth_type == self.T_AUTH_EXT_BROWSER:
                session = self._do_externalbrowser_auth(self.connect_params)
            elif self.auth_type == self.T_AUTH_OAUTH:
                session = self._do_oauth(self.connect_params, ctx)
        return session
    

    def close_session(self):
        """
        Close current snowflake connection and session.
        """
        self.close_connection()


    def _create_session(self, auth_type: str, conn_params) -> Session:
        try:
            session = Session.builder.configs(conn_params).create() 
            c = session.sql("select current_warehouse(), current_role()")\
                .collect()
            self.logger.info(
                "SnowConnect._create_session() Creation of connection [OK]"\
                f"Warehouse, Role => {c}"
            )
        except Exception as ex:
            self.logger.exception(
                "SnowConnect._create_session() Cannot connect to snowflake"\
                f" Error: {ex} with Auth_Type: {auth_type}"
            )
        return session


    def _do_snowflake_auth(self, params):
        pass_env: str = params["password_env"]
        password = os.environ[pass_env]
        conn_params = {
            "account": params["account"],
            "user": params["user"],
            "authenticator": self.T_AUTH_SNOWFLAKE,
            "password": password,
            "role": params["role"], 
            "warehouse": params["warehouse"],
            "database": params["database"],
            "schema": params["schema"]
        }
        return self._create_session(
            SnowConnect.T_AUTH_SNOWFLAKE, conn_params
        )


    def _do_keypair_auth(self, params):
        session, pkey, pkb, conn_params = None, None, bytes(), None
        kphrase_env = params["private_key_phrase_env"]
        kphrase = os.environ[kphrase_env]
        kpath_env = params["private_key_path_env"]
        kpath = os.environ[kpath_env]
        try:
            with open(kpath, "rb") as key:
                pkey = serialization.load_pem_private_key(
                    key.read(),
                    password = kphrase.encode(),
                    backend = default_backend()
                )
        except Exception as e:
            self.logger.exception(
                f"Cannot read snowflake user private key file: {e}"
            )
        if pkey:
            try:
                pkb = pkey.private_bytes(
                    encoding = serialization.Encoding.DER,
                    format = serialization.PrivateFormat.PKCS8,
                    encryption_algorithm = serialization.NoEncryption()
                )
            except Exception as e:
                self.logger.exception(
                    f"Cannot read snowflake user private key as bytes: {e}"
                )
        else: 
            raise ValueError(
                "Cannot read snowflake user private key file; "\
                f"check exception and key file path {kpath}."
            )
        if pkb:
            conn_params = {
                "account": params["account"],
                "user": params["user"],
                "private_key": pkb,
                "role": params["role"], 
                "warehouse": params["warehouse"],
                "database": params["database"],
                "schema": params["schema"],
                "client_session_keep_alive": True,
                "max_connection_pool" : 20
            }
            session = self._create_session(
                SnowConnect.T_AUTH_KEYPAIR, conn_params
            )
        else: 
            raise ValueError(
                "Cannot read snowflake user private key bytes in "\
                "SnowConnect._do_keypair_auth()"
            )
        return session


    def _do_externalbrowser_auth(self, params):
        conn_params = {
            "account": params["account"],
            "user": params["user"],
            "authenticator": self.T_AUTH_EXT_BROWSER,
            "role": params["role"], 
            "warehouse": params["warehouse"],
            "database": params["database"],
            "schema": params["schema"]
        }
        return self._create_session(
            SnowConnect.T_AUTH_EXT_BROWSER, conn_params
        )


    def _do_oauth(self, params, ctx: Dict):
        token = ctx.get("access_token")
        dtok = ctx.get("decoded_access_token")

        if dtok is not None and dtok:
            upn:str = dict(dtok).get("upn")
            upn = upn.upper()
        
            conn_params = {
                "account": params["account"],
                "user": upn,
                "authenticator": AppConnect.T_AUTH_OAUTH,
                "token" : token,
                "role": params["role"], 
                "warehouse": params["warehouse"],
                "database": params["database"],
                "schema": params["schema"]
            }
            return self._create_session(
                self.T_AUTH_OAUTH, conn_params
            )
    

    @staticmethod
    def ddl(session: Session, sql: str) -> List[Row]:
        """
        Excute Data Definition Language statement in Snowflake
        Args:
            sql (str): sql statement string

        Returns:
            List[Row]: list of Snowflake Snowpark Rows
        """
        cmds =  {
            'create', 'drop', 'alter', 'truncate', 'delete', 'replace',
            'rename', 'copy', 'clone', 'show', 'desc', 'undrop',
            'set', 'unset'
        }
        matches = [c for c in cmds if sql.upper().startswith(c.upper())]
        if not matches:
            raise ValueError(
                f"SnowConnect.ddl(): input sql doesn't seem to be DDL [{sql}]"
            )             
        return session.sql(sql).collect()


    @staticmethod
    def dml(session: Session, sql: str) -> List[Row]:
        """
        Excute Data Manupilation Language statement in Snowflake
        Args:
            sql (str): sql statement string

        Returns:
            List[Row]: list of Snowflake Snowpark Rows
        """
        cmds =  {'insert', 'update', 'delete', 'call', 'explain'}
        matches = [c for c in cmds if sql.upper().startswith(c.upper())]
        if not matches:
            raise ValueError(
                f"SnowConnect.dml(): input sql doesn't seem to be DML [{sql}]"
            )             
        return session.sql(sql).collect()


    @staticmethod    
    def tcl(session: Session, sql: str) -> List[Row]:
        """
        Excute Tranction Control Language statement in Snowflake
        Args:
            sql (str): sql statement string

        Returns:
            List[Row]: list of Snowflake Snowpark Rows
        """
        cmds =  {'begin', 'commit', 'rollback', 'savepoint', 'release'}
        matches = [c for c in cmds if sql.upper().startswith(c.upper())]
        if not matches:
            raise ValueError(
                f"SnowConnect.tcl(): input sql doesn't seem to be TCL [{sql}]"
            )             
        return session.sql(sql).collect()


    @staticmethod    
    def dcl(session: Session, sql: str) -> List[Row]:
        """
        Excute Data Control Language statement in Snowflake
        Args:
            sql (str): sql statement string

        Returns:
            List[Row]: list of Snowflake Snowpark Rows
        """
        cmds =  {
            'grant', 'revoke', 'create', 'drop', 'alter', 'use', 'show',
            'desc', 'set', 'unset'
        }
        matches = [c for c in cmds if sql.upper().startswith(c.upper())]
        if not matches:
            raise ValueError(
                f"SnowConnect.dcl(): input sql doesn't seem to be DCL [{sql}]"
            )             
        return session.sql(sql).collect()
    

    @staticmethod
    def dql(session: Session, sql: str) -> Iterator[DF]:
        """
        Excute Data Query Language (Select) statement in Snowflake
        Args:
            sql (str): sql statement string

        Returns:
            Iterator[DF]: Iterator of Pandas DataFrame list
        """
        cmds =  {'select'}
        matches = [c for c in cmds if sql.upper().startswith(c.upper())]
        if not matches:
            raise ValueError(
                f"SnowConnect.dql(): input sql doesn't seem to be DQL [{sql}]"
            )             
        return session.sql(sql).to_pandas_batches()
    

    @staticmethod
    def sdf(session: Session, sql: str) -> SDF:
        """
        Excute Data Query Language (Select) statement in Snowflake
        Args:
            sql (str): sql statement string

        Returns:
            Dataframe: Snowflake Dataframe as result of the query
        """
        cmds =  {'select'}
        matches = [c for c in cmds if sql.upper().startswith(c.upper())]
        if not matches:
            raise ValueError(
                f"SnowConnect.sdf(): input sql doesn't seem to be DQL [{sql}]"
            )
        return session.sql(sql).to_df()


# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains DataConnect class representing a generic data
connection configuration.
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.2.0"


import sys
from typing import Optional, Dict, Union
import logging

from snowflake.snowpark import Session
from snowflake_ai.common import AppConfig, AppConnect



class DataConnect(AppConnect):
    """
    This class represents a generic data connection.

    To use this class, create an instance from AppConfig or 
    extend it in a child class for creation of a specific data
    connection instance, e.g., SnowConnect as following :

        >>> from snowflake_ai.connect import SnowConnect
        ...
        ... # using default snowflake config directly
        >>> connect: SnowConnect = SnowConnect()
    """

    K_DATA_CONN = "data_connects"
    K_INIT_LIST = "init_list"

    _logger = logging.getLogger(__name__)
    _logger.addHandler(logging.StreamHandler(sys.stdout))

    _initialized = False

    # {<connect_key>: file-connect-string|snowflake-connection-object}
    _data_connections = {}


    def __init__(self, connect_key : Optional[str] = None):
        super().__init__(connect_key)
        self.logger = DataConnect._logger
        self._current_connection = None
        if (self.connect_key is not None) and self.connect_key:
            if ((self.data_connections.get(self.connect_key) is None) \
                    or (not self.is_current_active())) \
                    and (not self.is_oauth_saml_type()):
                self.data_connections[self.connect_key] = \
                    self.create_connection(self.connect_params)
                self.set_current_connection(
                    self.data_connections[self.connect_key]
                )

        if (connect_key is None or not connect_key or \
                self._current_connection is None) and self.data_connections:
            k = DataConnect.search_default_key(self.data_connections)
            self.connect_group, self.connect_key = \
                AppConfig.split_group_key(k)
            self.connect_key, self.connect_params = \
                AppConnect.load_connect_config(k, self.configs)
            self.app_connects[self.connect_key] = self
            self.set_current_connection(self.data_connections[k])

        if self.connect_params:
            self.connect_type = self.connect_params.get(AppConfig.K_TYPE, '')
            self.connect_name = self.connect_params.get(AppConfig.K_NAME, '')
            

    @property
    def data_connections(self) -> Dict:
        """
        Get a dictionary of all data connections.

        Returns:
            dict: a dictionary of all data connections.
        """   
        return DataConnect._data_connections


    def get_current_connection(self) -> Union[str, Session]:
        """
        Get current data connection.

        Returns:
            str, Session: specific data connection object, e.g. file
                path or Snowflake Session.
        """
        if not self.data_connections:
            self.logger.warning(
                "DataConnect.current_connection(): DataConnect."\
                "_data_connections dictionary is empty!"
            )
        elif (self._current_connection is None):
            k = DataConnect.search_default_key(self.data_connections)
            self.connect_group, self.connect_key = \
                AppConfig.split_group_key(k)
            self.connect_key, self.connect_params = \
                AppConnect.load_connect_config(k, self.configs)
            self.app_connects[self.connect_key] = self
            self.set_current_connection(self.data_connections[k])

        if ((not self.is_current_active()) or \
                (self.data_connections.get(self.connect_key) is None)) \
                and (not self.is_oauth_saml_type()):
            self.data_connections[self.connect_key] = \
                self.create_connection(self.connect_params)
            self.set_current_connection(
                self.data_connections[self.connect_key]
            )
        return self._current_connection
    

    def set_current_connection(self, conn: Union[str, Session]):
        """
        Set current data connection.

        Args:
            conn (str | Session): input connection object
        Returns:
            str, Session: specific data connection object, e.g. file
                path or Snowflake Session.
        """
        if conn is not None:
            self._current_connection = conn
        return self._current_connection


    def is_current_active(self) -> bool:
        """
        Check whether current data connection is active or not. It should
        be overwritten by specific subclass as it is implementation
        dependant.

        Returns:
            bool: True if it is active, otherwise False
        """
        if self._current_connection is None:
            return False
        elif not self._data_connections:
            return False
        elif self.is_oauth_saml_type():
            return False
        elif self.connect_type == AppConnect.T_SNOWFLAKE_CONN:
            return False
        return True
    

    def create_connection(self, params: Dict):
        """
        Create a data connection based on the specific configuration.
        Child class should implement specific connnection logic
        to override this method.

        Args:
            params (Dict): input parameters for connection creation.

        Returns:
            object: data connection object, e.g., connection session for
                snowflake connection.
        """
        conn = f"Connected to {params['host']}:{params['port']}"
        return conn


    def close_connection(self):
        """
        Close current data connection.
        
        Returns:
            int: 0 - successful; otherwise unsuccessful
        """
        rn = 0
        try:
            if self._current_connection is not None and \
                    isinstance(self._current_connection, Session):
                self._current_connection.close()
        except Exception as e:
            self.logger.error(f"DataConnect.close_connection(): Error {e}")
            rn = -1
        return rn


    def get_connection(self, connect_key: Optional[str] = None):
        """
        Get lazy created shared data connection or change existing data 
        connection to use the input connect_key to initialize connection.

        Args:
            conn_name (str): data connect name.

        Returns:
            object: data connection object, e.g., session connection for
                snowflake connection; it can return None.
        """
        if connect_key is None:
            if self._current_connection is None:
                if self.data_connections:
                    k = DataConnect.search_default_key(self.data_connections)
                    self.connect_group, self.connect_key = \
                        AppConfig.split_group_key(k)
                    self.connect_key, self.connect_params = \
                        AppConnect.load_connect_config(k, self.configs)
                    self.app_connects[self.connect_key] = self
                    self._current_connection = self.data_connections[k]
                else:
                    self.logger.warning(
                        f"DataConnect.get_connection(): connections "\
                        f"{len(self._data_connections)} initialized."
                    )
                    return None
            else:
                return self._current_connection
        else:
            qk = AppConfig.get_qualified_key(
                AppConfig.K_APP_CONN, connect_key
            )
            if self.data_connections.get(qk) is None:
                gk, k = AppConfig.split_group_key(qk)
                params = self.configs[AppConfig.K_APP_CONN]\
                    [DataConnect.K_DATA_CONN][k]
                self.connect_key = qk
                self.connect_params = params
                self.connect_group = gk
                if not self.is_oauth_saml_type():
                    self.data_connections[qk] = self.create_connection(params)
                    self._current_connection = self.data_connections[qk]

            return self.data_connections.get(qk)


    def init_connects(self) -> int:
        """
        Initialize DataConnect in [app_connects.data_connects] init_list.

        Returns:
            int: 0 - if there is no connect object has been initialized;
                otherwise, return an integer to show the number of 
                AppConnect objects have been initialized.
        """
        if not DataConnect._initialized:
            conn_group_dict = AppConfig.filter_group_key(
                DataConnect.K_DATA_CONN, 
                AppConnect.K_APP_CONN, 
                self.configs
            )
            lst = conn_group_dict.get(DataConnect.K_INIT_LIST)
            if lst is not None and len(lst) > 0 :
                for dconn in lst:
                    dc: dict = self.configs[AppConnect.K_APP_CONN]\
                        [DataConnect.K_DATA_CONN]
                    params = dc.get(dconn)
                    if params is None:
                        raise ValueError(
                            f"DataConnect.init_connets(): Error - [{dconn}]"\
                            " doesn't exist in the configuration!"
                        )
                    elif not self.is_oauth_saml_type():
                        c = self.create_connection(params)
                        s = f"{DataConnect.K_DATA_CONN}.{dconn}"
                        self.data_connections[s] = c
                DataConnect._initialized = True
        
        n = len(self.data_connections)
        self.logger.info(
            f"DataConnect.init_connects(): {n} non-oauth data connects."
        )
        return n



# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains AppConfig class representing application
configrations.
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.2.0"


import os
import sys
from os.path import exists
from pathlib import Path
from typing import Optional, Union, Dict, Tuple
import logging
import toml
from importlib.resources import read_text



class AppConfig:
    """
    This class represents the overall enterprise ai/ml application 
    configurations from a directory of configruation files and its
    instance represents a specific application configuration.

    The bootstrapping priority is the following: 

    1) input custom directory 
    2) snowflake_ai/conf subdir under current directory 
    3) snowflake_ai/conf subdir under user_home directory
    4) current directory
    5) home directory
    6) conf subdir under snowflake_ai library installation root dir

    Assuming there is a configruation of an application app_1 under a 
    default group named as group_0:

        >>> from snowflake_ai.common import AppConfig
        ... 
        ... # initialize application config for app_1 under group_0
        >>> ac = AppConfig("group_0.app_1")
        ...
        ... # optionally the configs may be loaded from your custom dir
        ... # ac = AppConfig("group_0.app_1", "custom dir", "config.toml")
        ...
        ... # get the default application which is streamlit app
        >>> app = ac.create_app()
    """

    DEFAULT_CONF_LIB_PATH = "snowflake_ai.conf"
    DEFAULT_CONF_FILE = "app_config.toml"
    DEFAULT_CONF_DIR = "./snowflake_ai/conf/"
    DEFAULT_CONN = "snowflake_0"
    DEFAULT_CURR_PATH = "."
    DEFAULT_HOME_PATH = "~"    

    T_SNOWFLAKE_CONN = "snowflake"
    T_FILE_CONN = "file"    
    T_ST_ML_APP = "streamlit_ml"
    T_OAUTH = "oauth"

    K_APPS = "apps"
    K_APP_CONN = "app_connects"
    K_DATA_SETUPS = "data_setups"
    K_ST_APPS = "streamlit_apps"
    K_MODEL_DEVS = "model_devs"
    K_MLOPS = "mlops"

    K_DEFAULT = "default"
    K_NAME = "name"
    K_TYPE = "type"
    K_RT_PATH = "root_path"
    K_VERSION = "version"
    K_ENV = "environment"
    K_AUTH_TYPE = "auth_type"

    K_OAUTH_CONN = "oauth_connects"
    K_MLOPS_PPL = "mlops_pipeline"

    RTK_LST = [
        K_APPS, K_APP_CONN, K_DATA_SETUPS, K_ST_APPS,
        K_MODEL_DEVS, K_MLOPS, 
    ]

    logging.basicConfig(
        level=logging.ERROR,
        format='%(asctime)s [%(levelname)s]  %(message)s'
    )
    _logger = logging.getLogger(__name__)
    _logger.addHandler(logging.StreamHandler(sys.stdout))

    _initialized = False
    _configs_root = ''
    _configs_path = ''
    _configs = {}
    _apps = {}


    def __init__(
        self,
        app_key : str,
        config_dir : Optional[Union[str, None]] = None, 
        config_file : Optional[Union[str, None]] = None
    ):
        """
        Creat app specific AppConfig object.

        Args:
            app_key: A string representing the application; it can have 
                the format of <app_group>.<application> ('<', '>' not 
                included)
            config_dir: directory path string for custom config load
            config_file: file name string for custom config load
        """
        self.logger = AppConfig._logger

        self.app_key = app_key.strip().lower()
        self.root_path = AppConfig._init_all_configs(config_dir, config_file)
        self.app_key, self.app_config = AppConfig.load_app_config(self.app_key)
        self.type =  self.get_all_configs()[AppConfig.K_APPS].get(
            AppConfig.K_TYPE, AppConfig.T_ST_ML_APP
        )
        self.root_path =  self.get_all_configs()[AppConfig.K_APPS].get(
            AppConfig.K_RT_PATH, os.path.abspath(AppConfig.DEFAULT_CURR_PATH)
        )
        self._init_app_base_config()
        if bool(self.app_key) and bool(self.app_config):
            self.apps[self.app_key] = self

        self.app_connect_keys = self.app_config.get(AppConfig.K_APP_CONN, [])
        self.data_setup_keys = self.app_config.get(AppConfig.K_DATA_SETUPS , [])


    @staticmethod
    def load_configs(
        config_dir : Optional[Union[str, None]] = None, 
        config_file : Optional[Union[str, None]] = None
    ) -> Tuple[str, str, Dict]:
        """
        Load configurations from input configuration directory and file if 
        supplied, or load them based on bootstrapping priorities.

        Args:
            config_dir (str): Configuration files directory.
            config_file (str): Configuration file name. If it is none or 
                empty, merging all toml files to create a complete config
                dictionary.

        Returns:
            Tuple[str, str, Dict]: root dir path and dir path from where
                the configurations are loaded, and the dictionary of all
                loaded configurations.

        Raises:
            ValueError: If config root, config file or path doesn't exist.
        """
        config_rt = None
        if (config_dir is None) or (not config_dir):
            config_rt = AppConfig.DEFAULT_CURR_PATH
            config_dir = AppConfig.DEFAULT_CONF_DIR
        configs = AppConfig._load_toml_files(config_dir, config_file)
        if config_rt != AppConfig.DEFAULT_CURR_PATH:
            config_rt = config_dir
        if config_rt:
            config_rt = os.path.abspath(config_rt)
        AppConfig._logger.info(
            f"AppConfig.load_configs(): 1:[config_root : {config_rt}];"\
            f" 1:[config_dir : {config_dir}"
        )

        home = Path.home()
        if not configs:
            config_rt = os.path.abspath(home)
            config_dir = os.path.join(home, AppConfig.DEFAULT_CONF_DIR)
            configs = AppConfig._load_toml_files(config_dir, config_file)
            AppConfig._logger.info(
                f"AppConfig.load_configs(): 2:[config_root : {config_rt}];"\
                f" 2:[config_dir : {config_dir}]"
            )
        
        if not configs:
            config_rt = os.path.abspath(AppConfig.DEFAULT_CURR_PATH)
            config_dir = config_rt
            configs = AppConfig._load_toml_files(config_dir, config_file)
            AppConfig._logger.info(
                f"AppConfig.load_configs(): 3:[config_root : {config_rt}];"\
                f" 3:[config_dir : {config_dir}]"
            )
        
        if not configs:
            config_rt = os.path.abspath(home)
            config_dir = config_rt
            configs = AppConfig._load_toml_files(config_dir, config_file)        
            AppConfig._logger.info(
                f"AppConfig.load_configs(): 4:[config_root : {config_rt}];"\
                f" 4:[config_dir : {config_dir}]"
            )
        
        if not configs:
            config_dir = AppConfig.DEFAULT_CONF_LIB_PATH
            configs = AppConfig.load_default_configs()
            AppConfig._logger.info(
                f"AppConfig.load_configs(): 5:[config_root : {config_rt}];"\
                f" 5:[config_dir : {config_dir}]"
            )
        
        if (config_rt is not None) and config_rt:
            config_rt =  os.path.abspath(config_rt)
            AppConfig._logger.info(
                f"AppConfig.load_configs(): 6:[config_root : {config_rt}];"\
                f" 6:[config_dir : {config_dir}]"
            )
        
        if (config_rt is None) or (not exists(config_rt)):
            s = f"AppConfig.load_configs(): Error - [{config_rt}] "\
                "doesn't exist!"
            AppConfig._logger.error(s)
            raise ValueError(s)

        return (config_rt, config_dir, configs)


    @staticmethod
    def _load_toml_files(
        config_dir : str, 
        config_file : Optional[str] = None
    ) -> Dict:
        rd = {}
        files_tsd = {}
        if (config_dir is not None) and (config_dir.strip()):
            config_dir = os.path.abspath(config_dir.strip())

        if not exists(config_dir):
            AppConfig._logger.warning(
                f"AppConfig._load_toml_files() directory {config_dir} "\
                "doesn't exist"
            )
            return rd

        if (config_file is not None) and bool(config_file.strip()):
            config_file_path = os.path.join(config_dir, config_file)
            with open(config_file_path, 'r') as f:
                try:
                    rd = toml.load(f)
                except Exception as e:
                    raise ValueError(
                        f"AppConfig._load_toml_files(): Cannot load "\
                        f"{config_file_path}, check format. Error - {e}"
                    )
            return rd
        else:
            toml_files = [
                f for f in os.listdir(config_dir) \
                    if f.lower().endswith(".toml")
            ]
            for toml_file in toml_files:
                config_file_path = os.path.join(config_dir, toml_file)
                with open(config_file_path, 'r') as f:
                    try:
                        toml_dict = toml.load(f)
                    except Exception as e:
                        raise ValueError(
                            f"AppConfig._load_toml_files(): Cannot load "\
                            f"files[{config_file_path}], please check format!"\
                            f" Error - {e}."
                        )
                file_ts = os.path.getmtime(config_file_path)

                for key, value in toml_dict.items():
                    key = key.strip().lower()
                    if key not in rd or file_ts > files_tsd[key]:
                        rd[key] = value
                        files_tsd[key] = file_ts
        
        AppConfig._logger.debug(
            f"DataConnect._load_toml_files(): From Dir => "\
            f"{config_dir}. Loaded Configs => {rd}"
        )
        return rd


    @staticmethod
    def load_default_configs() -> Dict:
        """
        Load default configuration from libary.

        Returns:
            dict: loaded configuration as a dictionary
        """
        config_file = read_text(
            AppConfig.DEFAULT_CONF_LIB_PATH, 
            AppConfig.DEFAULT_CONF_FILE
        )   
        configs = toml.loads(config_file)
        AppConfig._logger.info(
            f"Default app configuration loaded => {configs.keys}"
        )
        return configs
    

    @staticmethod
    def load_app_config(
        app_key: str, 
        configs: Optional[Union[Dict, None]] = None
    ) -> Tuple[str, Dict]:
        """
        Load app configuration from overall configurations

        Args:
            app_key (str): application key.
            configs (Dict): overall configuration dictionary

        Returns:
            Tuple[str, dict]: tuple of the app key string matched in
            a form of group.app_key and the dictionary of loaded
            application configurations
        """
        if configs is not None and configs.get(AppConfig.K_APPS) is None:
            s = f"AppConfig.load_app_config() Error: [apps] is missing"
            AppConfig._logger.error(s)
            raise ValueError(s)
        if app_key is None:
            return ('', {})
        
        if configs is None:
            configs = AppConfig.get_all_configs()

        rd = {}
        app_key = app_key.strip().lower()
        gk, ak = AppConfig.split_group_key(app_key)
        if not gk and bool(ak):
            k, rd = AppConfig.search_key_by_group(
                ak, AppConfig.K_APPS, configs
            )
        elif not gk and not ak:
            k, rd = '', {}
        elif gk and not ak:
            k, rd = f"{gk}.", AppConfig.filter_group_key(
                gk, AppConfig.K_APPS, configs
            )
        else:
            gs = dict(configs[AppConfig.K_APPS]).get(gk)
            if gs is not None:
                k = f"{gk}.{ak}" 
                rd =  configs[AppConfig.K_APPS][gk][ak] \
                    if dict(configs[AppConfig.K_APPS][gk]).get(ak) \
                        is not None else {}
            else:
                k, rd =  f"{gk}.{ak}", {}

        AppConfig._logger.info(
            f"AppConfig.load_app_config(): App[{k}] => {rd}"
        )
        return (k, rd)


    @staticmethod
    def filter_group_key(group_key: str, top_key:str, configs: Dict) \
        -> Dict:
        """
        Filter group section of configuration dictionary which has the
        structure of top_key: { group_key : { key1: "value1", key2: {} } }
        so that the returned dictionary group section doesn't have any
        nested dictionaries or subsections, i.e., {key1: "value1"} is
        returned as an example.

        Args:
            group_key (str): A string representing the group key
            top_key (str): A string that is the key of parent of group
                sections as the top level.
            configs (Dict): Configuration dictionary

        Returns:
            A dict containing group section's simple key and value pairs
            with any subsections filtered out.
        """
        if top_key not in configs or group_key not in configs[top_key]:
            return {}

        group_dict: dict = configs[top_key][group_key]

        filtered_dict = {k: v for k, v in group_dict.items() \
            if not isinstance(v, dict)}

        return filtered_dict


    @staticmethod
    def search_key_by_group(key: str, top_key:str, configs: Dict) \
        -> Tuple[str, Dict]:
        """
        Search sections of configs dictionary which has the structure of 
        top_key: { group_key: { key: {} } }, and return a tuple as 
        (group.key, dict of matched key section).

        Args:
            key (str): A string representing the input key to be searched.
            top_key (str): A string that is the key of parent of group
                sections as the top level.
            configs (Dict): Configuration dictionary

        Returns:
            A tuple containing the group_key.key and dict of matched
            key section.
        """
        if top_key not in configs:
            return '',  {}

        default_match = '',  {}
        first_match = '',  {}

        config_groups :dict = configs[top_key]
        for group_key in sorted(config_groups.keys()):
            group_dict = config_groups[group_key]
            if key in group_dict:
                if AppConfig.K_DEFAULT in group_key:
                    default_match = f"{group_key}.{key}", group_dict[key]
                elif first_match == ('', {}):
                    first_match = f"{group_key}.{key}", group_dict[key]

        if default_match != ('', {}):
            return default_match
        else:
            return first_match


    @staticmethod
    def split_group_key(key: str) -> Tuple[str, str]:
        """
        Split an input key by '.' and return group_key and the key without
        group prefix.

        Args:
            key: A string representing the input key to be split.

        Returns:
            A tuple containing the group_key and key without group prefix
            as strings.
        """
        if key is None:
            return '', ''
        key = key.strip().lower()
        split_keys = key.split('.')
        if len(split_keys) > 1:
            group_k = split_keys[0]
            k = split_keys[1]
        else:
            k = split_keys[0]
            group_k = ''
        return (group_k, k)


    @classmethod
    def get_qualified_key(cls, root_key: str, key: str) -> str:
        """
        Get fully qualified key in form of <group_key>.<key>

        Returns:
            A string representing a fully qualified key in config dict.
        """
        if root_key not in AppConfig.RTK_LST:
            raise ValueError(
                f"AppConfig.get_qualified_key(): Error - [{key}] not found!"
            )
        
        gk, k = AppConfig.split_group_key(key)
        if k:
            configs: dict = cls.get_all_configs()
            ret_k, _ = AppConfig.search_key_by_group(k, root_key, configs)
            if not ret_k:
                raise ValueError(
                    f"AppConfig.get_qualified_key(): Error - Key Empty!"
                )
            else:
                g_dict = dict(configs[root_key]).get(gk)
                if g_dict is not None:
                    ret_k = f"{gk}.{k}"
        else:
            raise ValueError(
                f"AppConfig.get_qualified_key(): Error - Input [{key}]"
            )
        return ret_k


    @classmethod
    def _init_all_configs(
        cls,
        config_dir : Optional[Union[str, None]] = None, 
        config_file : Optional[Union[str, None]] = None
    ) -> str:
        """
        Initialize the dictionary of all application configurations

        Returns:
            The root directory path where the configs are loaded from.
        """        
        if not cls._initialized:
            cls._configs_root, cls._configs_path, cls._configs = \
                AppConfig.load_configs(config_dir, config_file)
            cls._initialized = True
        return cls._configs_root


    def _init_app_base_config(self):
        self.name = self.app_config.get(AppConfig.K_NAME)
        self.group = AppConfig.split_group_key(self.app_key)[0]
        s = self.app_config.get(AppConfig.K_TYPE)
        if s is not None and s:
            self.type = s
        elif self.type is None or not self.type:
            self.type = AppConfig.T_ST_ML_APP
        self.version = self.app_config.get(AppConfig.K_VERSION)
        self.environment= self.app_config.get(AppConfig.K_ENV)
        s = self.app_config.get(AppConfig.K_RT_PATH)
        if s is not None and s:
            self.root_path = s
        elif self.root_path is None or not self.root_path:
            self.root_path = os.path.abspath(AppConfig.DEFAULT_CURR_PATH)
        self.app_dir = self.app_config.get("app_dir", ".")
        self.script_home_dir = self.app_config.get("script_home_dir", ".")
        self.group_config = AppConfig.filter_group_key(
            self.group, AppConfig.K_APPS, self.get_all_configs()
        )
    

    @classmethod
    def get_configs_root(cls) -> str:
        """
        Get root path of directory where configurations are loaded.

        Returns:
            str: root path string of directory containing all 
                configuration files.
        """
        if not cls._configs_root:
            cls._configs_root = cls._init_all_configs()
        return cls._configs_root
    

    @classmethod
    def get_configs_path(cls) -> str:
        """
        Get path of directory where configurations are loaded.

        Returns:
            str: path string of directory containing all 
                configuration files.
        """
        if not cls._configs_path:
            cls._init_all_configs()
        return cls._configs_path
    

    @classmethod
    def get_all_configs(cls) -> Dict[str, Dict]:
        """
        Get overall configurations.

        Returns:
            dict: all configurations.
        """ 
        if not cls._configs:
            cls._init_all_configs()
        return cls._configs


    @classmethod
    def get_all_apps(cls) -> Dict:
        """
        Get all initialized applications.

        Returns:
            dict: all initialized applications.
        """  
        return cls._apps


    @property
    def apps(self) -> Dict:
        """
        Get/set a dictionary of all initialized applications.

        Returns:
            dict: a dictionary of all initialized applications.
        """   
        return AppConfig._apps
    

    def get_app_type(self) -> str:
        """
        Get application configured type. Currently supported type
        is "streamlit_ml"

        Returns:
            str: application type string.
        """   
        return self.type


    def get_root_path(self) -> str:
        """
        Get overall apps configuration root path

        Returns:
            str: root directory path for all applications
        """   
        return self.root_path


    def get_app_dir(self) -> str:
        """
        Get the directory of the specified application

        Returns:
            str: the directory of the specified application
        """   
        return self.app_dir
    



# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains Pipeline class for ML flow focusing on model
building, training, validation and testing in DEV domain.
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.2.0"


import logging

from snowflake_ai.mlops import FlowContext


# Configure logging
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s [%(levelname)s]  %(message)s'
)



class Pipeline:
    """
    This class provides general setup for MLops flow in
    model building, training, validation and testing. 
    (TO-DO).
    """
    
    def __init__(self) -> None:
        self.logger = logging.getLogger(__name__)
        self.flow_context = FlowContext()

    @classmethod
    def flow(cls, func):
        def wrapper(
            obj: Pipeline
        ):
            _logger = logging.getLogger(__name__)
            ctx = FlowContext()
            if obj.flow_context is None:
                obj.flow_context = ctx
            else:
                ctx = obj.flow_context

            if ctx.direct_input is None:
                _logger.warn("FlowContext direct input is None")
                ctx.direct_input = {}
                return obj
            elif ctx.context_input is None:
                _logger.warn("FlowContext context input is None")
                ctx.context_input = []
                return obj
            else:
                if ctx.direct_input and ctx.context_input:
                    ctx.context_input.append(ctx.direct_input)
                    func(obj)
                    ctx.context_input.append(ctx.output)
                    return obj
                elif ctx.context_input:
                    func(obj)
                    ctx.context_input.append(ctx.output)
                    return obj
                else:
                    ctx.context_input = []
                    func(obj)
                    ctx.context_input.append(ctx.output)
                    return obj
        return wrapper
    

    def run(self):
        return 0



# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains OAuthConnect class representing a generic OAuth
connection configuration.
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.2.0"


import sys
from typing import Optional, Dict, List
import logging

from snowflake_ai.common import AppConfig, OAuthConnect
from snowflake_ai.common import AppConnect, DataConnect
from snowflake_ai.connect import SnowConnect, AuthCodeConnect


class ConnectManager:
    """
    This class creates and manages application connects.

    To use this class, create instance from AppConfig or extend
    it in a child class for creation of a specific OAuth connect
    instance, e.g., AuthCodeConnect as following :

        >>> from snowflake_ai.connect import AuthCodeConnect
        ...
        >>> ac = AppConfig("group_0.app_1")
        >>> conn = SecurityManager(ac).create_default_oauth_connect()
        >>> conn.grant_request()
    """

    _logger = logging.getLogger(__name__)
    _logger.addHandler(logging.StreamHandler(sys.stdout))

    _initialized = False
    _app_connect_list = []


    @classmethod
    def create_default_oauth_connect(cls, app_config: AppConfig) \
            -> OAuthConnect:
        """
        Create default oauth connect.

        Returns:
            OAuthConnect: default oauth connect
        """   
        if not cls._initialized:
            cls.create_app_connects(app_config)
        for c in cls._app_connect_list:
            if isinstance(c, OAuthConnect):
                return c
        return OAuthConnect()
    

    @classmethod
    def create_default_snow_connect(cls, app_config: AppConfig) \
            -> SnowConnect:
        """
        Create default snowflake connect.

        Returns:
            OAuthConnect: default snowflake connect
        """   
        if not cls._initialized:
            cls.create_app_connects(app_config)
        for c in cls._app_connect_list:
            if isinstance(c, SnowConnect):
                return c
        return SnowConnect()
    

    @classmethod
    def create_app_connects(cls, app_config: AppConfig) -> List:
        """
        Create list of application connect instances

        Returns:
            List: application connect instances
        """ 
        if app_config is None:
            raise ValueError(
                "ConnectManager.create_app_connects(): AppConfig is required!"
            )
        if not cls._initialized:
            for ck in app_config.app_connect_keys:
                gk, k = AppConfig.split_group_key(ck)
                config = app_config.get_all_configs()\
                    [AppConfig.K_APP_CONN][gk][k]
                
                if gk == OAuthConnect.K_OAUTH_CONN:
                    if config[AppConfig.K_TYPE] == "auth_code":
                        cls. _app_connect_list.append(AuthCodeConnect(ck))
                        print(f"DEBUG C1 => {config[AppConfig.K_TYPE] }")
                elif gk == DataConnect.K_DATA_CONN:
                    if config[AppConfig.K_TYPE] == "snowflake":
                        cls. _app_connect_list.append(SnowConnect(ck))
                        print(f"DEBUG C2 => {config[AppConfig.K_TYPE] }")
            
            cls._initialized = True

        return cls. _app_connect_list


# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains AppConnect class representing a generic
application integration connection for data, security, etc.
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.2.0"


import sys
from typing import Optional, Dict, Union, Tuple
import logging

from snowflake_ai.common import AppConfig



class AppConnect:
    """
    This class represents a generic application connection.
    Currently it is extended for oauth_connect, and
    data_connect.

    To use this class, create specific instances from AppConfig
    or extend it in a child class such as DataConnect. 
    """

    K_OAUTH_CONN = "oauth_connects"
    K_DATA_CONN = "data_connects"
    
    T_SNOWFLAKE_CONN = AppConfig.T_SNOWFLAKE_CONN
    T_AUTH_SNOWFLAKE = "snowflake"
    T_AUTH_KEYPAIR = "keypair"
    T_AUTH_EXT_BROWSER = "externalbrowser"
    T_AUTH_OAUTH = "oauth"

    SUPPORTED_CONNECT_GROUP_TYPES = [
        K_OAUTH_CONN, 
        K_DATA_CONN
    ]
    K_APP_CONN = AppConfig.K_APP_CONN

    _logger = logging.getLogger(__name__)
    _logger.addHandler(logging.StreamHandler(sys.stdout))

    _configs = AppConfig.get_all_configs()
    _connects = {}


    def __init__(self, connect_key : Optional[str] = None):
        """
        Creat an AppConnect object.

        Args:
            connect_key: A string representing an AppConnect object;
                it can have the format of <app_connect_group>.<app_
                connect> ('<', '>' not included)
        """
        self.logger = AppConnect._logger
        self.connect_type, self.connect_name, self.auth_type = '', '', ''
        
        self.connect_group, self.connect_key = \
            AppConfig.split_group_key(connect_key)
        self.connect_key, self.connect_params = \
            AppConnect.load_connect_config(self.connect_key, self.configs)
        self.logger.info(
            f"AppConnect.init(): ConnectParameters => {self.connect_params}"
        )

        if self.connect_key and self.connect_params:
            self.app_connects[self.connect_key] = self

        if self.connect_params:
            self.auth_type = self.connect_params.get(
                AppConfig.K_AUTH_TYPE, ''
            )

        # setup initialization
        self.init_connects()


    @property
    def configs(self) -> Dict:
        """
        Get all configurations loaded from AppConfig.

        Returns:
            dict: a dictionary of all configurations loaded.        
        """
        return AppConnect._configs
    

    @staticmethod
    def load_connect_config(
        connect_key: str, 
        configs: Optional[Union[Dict, None]] = None
    ) -> Tuple[str, Dict]:
        """
        Load app connect configuration from overall configurations

        Args:
            connect_key (str): app connect key.
            configs (Dict): overall configuration dictionary

        Returns:
            Tuple[str, dict]: tuple of the connect key string matched in
            a form of group.connect_key and the dictionary of loaded
            app connect specific configurations
        """
        if configs is not None and \
            configs.get(AppConnect.K_APP_CONN) is None:
            s = f"AppConnect.load_connect_config(): Error - "\
                "[app_connects] is missing!"
            AppConnect._logger.error(s)
            raise ValueError(s)
        if connect_key is None:
            return ('', {})
        
        if configs is None:
            configs = AppConfig.get_all_configs()

        rd = {}
        connect_key = connect_key.strip().lower()
        gk, ck = AppConfig.split_group_key(connect_key)
        if not gk and ck:
            k, rd = AppConfig.search_key_by_group(
                    ck, AppConnect.K_APP_CONN, configs
                )
        elif not gk and not ck:
            k, rd = '', {}
        elif gk and not ck:
            k, rd = f"{gk}.", AppConfig.filter_group_key(
                    gk, AppConnect.K_APP_CONN, configs
                )
        else:
            gs = dict(configs[AppConnect.K_APP_CONN]).get(gk)
            if gs is not None:
                k = f"{gk}.{ck}" 
                rd =  configs[AppConnect.K_APP_CONN][gk][ck] if dict(
                        configs[AppConnect.K_APP_CONN][gk]
                    ).get(ck) is not None else {}
            else:
                k, rd =  f"{gk}.{ck}", {}

        AppConfig._logger.info(
            f"AppConnect.load_connect_config(): AppConnect[{k}] => {rd}"
        )
        return (k, rd)


    @staticmethod
    def search_default_key(data_dict: Dict[str, object]) -> str:
        """
        Search a dictionary for a key containing the substring "default". 
        If found, return that key. If not found, check if "_0", return
        the first found. If still not found any, sort the keys and return
        the first key in the sorted list.

        Args:
            data_dict (dict): A dictionary with string keys.

        Returns:
            str, None: A key containing "default" or "_0". If found, or the
            first key in the sorted list if not found.
        """
        if not data_dict: return None

        default_key = None
        sorted_keys = sorted(data_dict.keys())

        for k in sorted_keys:
            if AppConfig.K_DEFAULT in k.lower():
                default_key = k
                break

            if "_0" in k:
                default_key = k
                break

        return default_key or sorted_keys[0]
    

    @property
    def app_connects(self) -> Dict:
        """
        Get/set a dictionary of all currently loaded application connections.

        Returns:
            dict: a dictionary of all application connections.
        """   
        return AppConnect._connects
    

    def init_connects(self) -> int:
        """
        Initialize default AppConnect in each group. This generally should
        be overwritten by its child class
        """
        return len(AppConnect._connects)


    def is_oauth_saml_type(self) -> bool:
        """
        Return whether this application connect has auth_type == "oauth"
        or auth_type == "externalbrowser"
        """
        auth_t = self.auth_type
        return (auth_t == AppConfig.T_OAUTH) or \
                (auth_t == AppConnect.T_AUTH_EXT_BROWSER)
